#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import datetime
from typing import Literal, List, Optional
from xml.etree.ElementTree import Element

from Annotations2Sub.Color import Alpha, Color
from Annotations2Sub.locales import _


class Annotation(object):
    # 致谢 https://github.com/isaackd/annotationlib
    """Annotation 结构"""

    def __init__(self):
        # 仅列出了需要的结构
        self.id: str = ""
        # 这里仅列出需要的的 type 和 style
        self.type: Literal["text", "highlight", "branding"] = ""
        self.style: Literal[
            "popup",
            "title",
            "speech",
            "highlightText",
        ] = ""
        self.text: str = ""
        self.timeStart: datetime.datetime = datetime.datetime.strptime("0", "%S")
        self.timeEnd: datetime.datetime = datetime.datetime.strptime("0", "%S")
        self.x: float = 0.0
        self.y: float = 0.0
        self.width: float = 0.0
        self.height: float = 0.0
        # sx, sy 是气泡锚点
        self.sx: float = 0.0
        self.sy: float = 0.0
        self.bgOpacity: Alpha = Alpha(alpha=204)
        self.bgColor: Color = Color(red=255, green=255, blue=255)
        self.fgColor: Color = Color(red=0, green=0, blue=0)
        self.textSize: float = 3.15


def Parse(tree: Element) -> List[Annotation]:
    """XML 树转换为 Annotation 结构列表"""

    def ParseAnnotationAlpha(annotation_alpha_str: str) -> Alpha:
        """
        解析 Annotation 的透明度
        bgAlpha="0.600000023842" -> Alpha(alpha=102)
        """
        s0 = annotation_alpha_str
        if s0 == None:
            raise Exception("alpha is None")
        s1 = float(s0)
        s2 = 1 - s1
        s3 = s2 * 255
        s4 = int(s3)
        s5 = Alpha(alpha=s4)
        return s5

    def ParseAnnotationColor(annotation_color_str: str) -> Color:
        """
        bgColor="4210330" -> Color(red=154, green=62, blue=64)
        """
        s0 = annotation_color_str
        if s0 == None:
            raise Exception("color is None")
        s1 = int(s0)
        r = s1 & 255
        g = (s1 >> 8) & 255
        b = s1 >> 16
        s2 = Color(red=r, green=g, blue=b)
        return s2

    def MakeSureStr(s: Optional[str]) -> str:
        if isinstance(s, str):
            return str(s)
        raise TypeError

    def ParseAnnotation(each: Element) -> Optional[Annotation]:
        # 致谢: https://github.com/nirbheek/youtube-ass
        # 致谢: https://github.com/isaackd/annotationlib
        annotation = Annotation()

        annotation.id = MakeSureStr(each.get("id"))

        type = each.get("type")
        if type not in ("text", "highlight", "branding"):
            print(_("本脚本暂不支持{}类型. ()").format(type, annotation.id))
            return None
        annotation.type = MakeSureStr(type)  # type: ignore

        annotation.style = each.get("style")  # type: ignore

        text = each.find("TEXT")
        if text is None:
            annotation.text = ""
        else:
            annotation.text = MakeSureStr(text.text)

        if len(each.find("segment").find("movingRegion")) == 0:  # type: ignore
            # 跳过没有内容的 Annotation
            return None

        Segment = each.find("segment").find("movingRegion").findall("rectRegion")  # type: ignore
        if len(Segment) == 0:
            Segment = (
                each.find("segment").find("movingRegion").findall("anchoredRegion")  # type: ignore
            )

        if len(Segment) == 0:
            if annotation.style != "highlightText":
                # 抄自 https://github.com/isaackd/annotationlib/blob/master/src/parser/index.js 第121行
                # "highlightText" 是一直显示在屏幕上的, 不应没有时间
                return None

        if len(Segment) != 0:
            t1 = MakeSureStr(Segment[0].get("t"))
            t2 = MakeSureStr(Segment[1].get("t"))
            Start = min(t1, t2)
            End = max(t1, t2)

        if "never" in (Start, End):
            # 跳过不显示的 Annotation
            return None

        try:
            annotation.timeStart = datetime.datetime.strptime(Start, "%H:%M:%S.%f")
            annotation.timeEnd = datetime.datetime.strptime(End, "%H:%M:%S.%f")
        except:
            annotation.timeStart = datetime.datetime.strptime(Start, "%M:%S.%f")
            annotation.timeEnd = datetime.datetime.strptime(End, "%M:%S.%f")

        annotation.x = float(MakeSureStr(Segment[0].get("x")))
        annotation.y = float(MakeSureStr(Segment[0].get("y")))

        w = Segment[0].get("w")
        h = Segment[0].get("h")
        sx = Segment[0].get("sx")
        sy = Segment[0].get("sy")

        if w is not None:
            annotation.width = float(MakeSureStr(w))
        if h is not None:
            annotation.height = float(MakeSureStr(h))
        if sx is not None:
            annotation.sx = float(MakeSureStr(sx))
        if sy is not None:
            annotation.sy = float(MakeSureStr(sy))

        Appearance = each.find("appearance")

        if Appearance != None:
            bgAlpha = Appearance.get("bgAlpha")  # type: ignore
            bgColor = Appearance.get("bgColor")  # type: ignore
            fgColor = Appearance.get("fgColor")  # type: ignore
            textSize = Appearance.get("textSize")  # type: ignore

        if bgAlpha != None:
            annotation.bgOpacity = ParseAnnotationAlpha(MakeSureStr(bgAlpha))
        if bgColor != None:
            annotation.bgColor = ParseAnnotationColor(MakeSureStr(bgColor))
        if fgColor != None:
            annotation.fgColor = ParseAnnotationColor(MakeSureStr(fgColor))
        if textSize != None:
            annotation.textSize = float(MakeSureStr(textSize))

        return annotation

    annotations: List[Annotation] = []
    for each in tree.find("annotations").findall("annotation"):  # type: ignore
        annotation = ParseAnnotation(each)
        if annotation != None:
            annotations.append(annotation)  # type: ignore

    return annotations



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Annotations2Sub, 一个能把 Youtube 注释转换成 Sub Station Alpha V4 字幕文件的脚本

                                            xml.
                                            etree.
┌────────────────┐            ┌────────┐    ElementTree.  ┌─────────┐
│                │    read()  │        │    fromstring()  │         │     Parse()
│ Annotation.xml │ ─────────► │ String │ ───────────────► │ Element │ ────────────┐
│                │            │        │                  │         │             │
└────────────────┘            └────────┘                  └─────────┘             │
                                                                                  ▼
                                                                         ┌──────────────────┐
                                                                         │                  │
                                                                         │ List[Annotation] │
                                                                         │                  │
                                                                         └──────────────────┘

┌────────────────┐            ┌────────┐            ┌─────────────┐               │
│                │  finally   │        │  Dump()    │             │  Convert()    │
│ Annotation.ass │ ◄───────── │ String │ ◄───────── │ List[Event] │ ◄─────────────┘
│                │            │        │            │             │
└────────────────┘            └────────┘            └─────────────┘

在 2019 年的某一天, 我在 Youtube 听歌, 发现字幕没了.
从评论区我得知: "從 2019 年 1 月 15 日起，我們將不再對觀眾顯示現有的註解。所有的現有註解都將移除。"
我当即破口大骂谷歌.
但我直觉告诉我应该有人存档,
还真的有! 鸣谢: https://archive.org/details/youtubeannotations
我想将其转换为我熟悉的 Sub Station Alpha 字幕, 我在搜索引擎查找我需要的东西, 找到了 https://github.com/nirbheek/youtube-ass (致谢)
但是他并不好用, 压根就没法用
我改呀改, 这时我其实并不会编程, 但是我会谷歌, 让他可以运行也算是没有花费太长的时间.
但是效果并不能让人满意, 没有颜色, 只有字幕, 很不满意
于是我仔细阅读了他的源码, 发现实现很恶心, 竟然用样式来实现定位, 而且并不是 Sub Station Alpha 目前的事实标准(Aegisub)
但是拜其所致我脑内有了大概的思路, 于是我觉得把他砸了重写一个 :-)
回过头来, 这个过程让我学会了编程.

此项目早期(8e2d8eb 第一个勉强可用的版本)参照 https://github.com/nirbheek/youtube-ass 实现了一个简单的转换器, 
参照 https://github.com/weizhenye/ASS/wiki/ASS-字幕格式规范 实现了一个简单的 Sub Station Alpha 格式生成器
使用 "样式覆写代码" 实现了定位和颜色
之后(a64995 庆祝一下), 使用了 "绘图模式" 实现了 "popup" 样式
在之后, 脚本逐步完善, 完成了 https://www.bilibili.com/video/BV1Ff4y1t7Dj

"""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sys
import os

if __name__ == "__main__":
    path = os.path.split(os.path.realpath(__file__))[0]
    sys.path.append(path)
    path = os.path.abspath(os.path.join('path', os.pardir))
    sys.path.append(path)
    from Annotations2Sub.cli import main

    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import argparse
import defusedxml.ElementTree  # type: ignore

from Annotations2Sub.Annotation import Parse
from Annotations2Sub.Convert import Convert
from Annotations2Sub.Sub import Sub
from Annotations2Sub.locales import _


def main():
    parser = argparse.ArgumentParser(
        description=_("一个可以把Youtube注释转换成ASS字幕(Sub Station Alpha V4)文件的脚本")
    )
    parser.add_argument(
        "File",
        type=str,
        nargs="+",
        metavar=_("文件 或 视频ID"),
        help=_("多个需要转换的文件或者是需要预览或生成 Youtube 视频的 videoId"),
    )
    args = parser.parse_args()
    file = args.File[0]
    string = open(file, "r", encoding="utf-8").read()
    tree = defusedxml.ElementTree.fromstring(string)
    annotations = Parse(tree)
    events = Convert(annotations)
    sub = Sub()
    sub.events.events.extend(events)
    sub.info.info["PlayResX"] = 100
    sub.info.info["PlayResY"] = 100
    sub.events.events.sort(key=lambda event: event.Start)
    sub.styles.styles["Default"].Fontname = "Microsoft YaHei"
    s = sub.Dump()
    with open(file + ".ass", "w", encoding="utf-8") as f:
        f.write(s)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import gettext
import os

locales = os.path.join(os.path.split(os.path.realpath(__file__))[0], "locales")
translate = gettext.translation(
    "Annotations2Sub",
    locales,
)
_ = translate.gettext



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
https://github.com/weizhenye/ASS/wiki/ASS-字幕格式规范
"""

import datetime
from typing import Dict, List, Literal

from Annotations2Sub.Color import Alpha, Color, Rgba


class Style:
    def __init__(self):
        self.Fontname: str = "Arial"
        self.Fontsize: float = 20
        self.PrimaryColour: Rgba = Rgba(Color(255, 255, 255), Alpha(0))
        self.SecondaryColour: Rgba = Rgba(Color(255, 0, 0), Alpha(0))
        self.OutlineColour: Rgba = Rgba(Color(0, 0, 0), Alpha(0))
        self.BackColour: Rgba = Rgba(Color(0, 0, 0), Alpha(0))
        self.Bold: Literal[-1, 0] = 0
        self.Italic: Literal[-1, 0] = 0
        self.Underline: Literal[-1, 0] = 0
        self.StrikeOut: Literal[-1, 0] = 0
        self.ScaleX: int = 100
        self.ScaleY: int = 100
        self.Spacing: int = 0
        self.Angle: float = 0
        self.BorderStyle: Literal[1, 3] = 1
        self.Outline: Literal[0, 1, 2, 3, 4] = 2
        self.Shadow: Literal[0, 1, 2, 3, 4] = 2
        self.Alignment: Literal[0, 1, 2, 3, 4, 5, 6, 7, 8, 9] = 2
        self.MarginL: int = 10
        self.MarginR: int = 10
        self.MarginV: int = 10
        self.Encoding: int = 1


class Event:
    """Sub 的 Event 结构"""

    def __init__(self):

        # 仅列出了需要的 Format
        self.Format: Literal["Dialogue"] = "Dialogue"
        # Aegisub 没有 Marked ,所以我们也没有
        self.Layer: int = 0
        self.Start: datetime.datetime = datetime.datetime.strptime("0", "%S")
        self.End: datetime.datetime = datetime.datetime.strptime("0", "%S")
        self.Style: str = "Default"
        self.Name: str = ""
        # MarginL, MarginR, MarginV, Effect 在本项目中均没有使用
        self.MarginL: int = 0
        self.MarginR: int = 0
        self.MarginV: int = 0
        self.Effect: str = ""
        self.Text: str = ""


class Point:
    """绘图命令"""

    def __init__(self, x=0, y=0, command="m"):
        self.x: int = x
        self.y: int = y
        # 仅列出需要的命令
        self.command: Literal["m", "l"] = command


class Draw:
    def __init__(self):
        self.draw: List[Point] = []

    def Add(self, point: Point):
        self.draw.append(point)

    def Dump(self) -> str:
        s = ""
        for i in self.draw:
            s = s + "{} {} {} ".format(i.command, i.x, i.y)
        return s


class Sub:
    def __init__(self):
        self.info = self.Info()
        self.styles = self.Styles()
        self.events = self.Events()

        self.info.info["Title"] = "Default File"
        self.info.note = (
            "; Script generated by Annotations2Sub" + "\n"
            "; https://github.com/USED255/Annotations2Sub" + "\n"
        )
        self.styles.styles["Default"] = Style()

    class Info:
        def __init__(self):
            self.note = ""
            self.info = {"ScriptType": "v4.00+"}

        def Dump(self) -> str:
            s = ""
            s += "[Script Info]" + "\n"
            s += self.note
            for k, v in self.info.items():
                s += "{}: {}\n".format(k, v)
            s += "\n"
            return s

    class Styles:
        def __init__(self):
            self.styles: Dict[str, Style] = {}

        def Dump(self) -> str:
            def ConvertRgbaToAABBGGRR(rgba: Rgba) -> str:
                return "&H{:02X}{:02X}{:02X}{:02X}".format(
                    rgba.alpha.alpha, rgba.color.blue, rgba.color.green, rgba.color.red
                )

            s = ""
            s += "[V4+ Styles]" + "\n"
            s += (
                "Format: Name, Fontname, Fontsize, PrimaryColour, SecondaryColour, OutlineColour, BackColour, Bold, Italic, Underline, StrikeOut, ScaleX, ScaleY, Spacing, Angle, BorderStyle, Outline, Shadow, Alignment, MarginL, MarginR, MarginV, Encoding"
                + "\n"
            )
            for name, each in self.styles.items():
                s += "Style: "
                s += name + ","
                s += each.Fontname + ","
                s += str(each.Fontsize) + ","
                s += ConvertRgbaToAABBGGRR(each.PrimaryColour) + ","
                s += ConvertRgbaToAABBGGRR(each.SecondaryColour) + ","
                s += ConvertRgbaToAABBGGRR(each.OutlineColour) + ","
                s += ConvertRgbaToAABBGGRR(each.BackColour) + ","
                s += str(each.Bold) + ","
                s += str(each.Italic) + ","
                s += str(each.Underline) + ","
                s += str(each.StrikeOut) + ","
                s += str(each.ScaleX) + ","
                s += str(each.ScaleY) + ","
                s += str(each.Spacing) + ","
                s += str(each.Angle) + ","
                s += str(each.BorderStyle) + ","
                s += str(each.Outline) + ","
                s += str(each.Shadow) + ","
                s += str(each.Alignment) + ","
                s += str(each.MarginL) + ","
                s += str(each.MarginR) + ","
                s += str(each.MarginV) + ","
                s += str(each.Encoding) + "\n"
            s += "\n"
            return s

    class Events:
        def __init__(self):
            self.events: List[Event] = []

        def Dump(self) -> str:
            def ConvertTime(t: datetime.datetime) -> str:
                return t.strftime("%H:%M:%S.%f")[:-4]

            s = ""
            s += "[Events]" + "\n"
            s += (
                "Format: Layer, Start, End, Style, Name, MarginL, MarginR, MarginV, Effect, Text"
                + "\n"
            )
            for each in self.events:
                s += "Dialogue: "
                s += str(each.Layer) + ","
                s += ConvertTime(each.Start) + ","
                s += ConvertTime(each.End) + ","
                s += each.Style + ","
                s += each.Name + ","
                s += str(each.MarginL) + ","
                s += str(each.MarginR) + ","
                s += str(each.MarginV) + ","
                s += each.Effect + ","
                s += each.Text + "\n"
            s += "\n"
            return s

    def Dump(self) -> str:
        s = ""
        s += self.info.Dump()
        s += self.styles.Dump()
        s += self.events.Dump()
        return s



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy
from typing import List

from Annotations2Sub.Annotation import Annotation
from Annotations2Sub.Color import Alpha, Color
from Annotations2Sub.Sub import Draw, Event, Point
from Annotations2Sub.locales import _


def Convert(annotations: List[Annotation], libass: bool = False) -> List[Event]:
    """将 Annotation 列表转换为 Event 列表"""

    def ConvertColor(color: Color) -> str:
        return "&H{:02X}{:02X}{:02X}&".format(color.red, color.green, color.blue)

    def ConvertAlpha(alpha: Alpha) -> str:
        return "&H{:02X}&".format(alpha.alpha)

    def ConvertAnnotation(each: Annotation) -> List[Event]:
        # 致谢: https://github.com/nirbheek/youtube-ass
        # 致谢: https://github.com/weizhenye/ASS/wiki/ASS-字幕格式规范
        def popup(event: Event) -> Event:
            event.Name += "_popup"

            tag = ""
            tag += r"\an7" + r"\pos({},{})".format(x + 1, y + 1)
            tag += r"\fs" + str(textSize)
            tag += r"\c" + fgColor
            tag += r"\2a" + "&HFF&" + r"\3a" + "&HFF&" + r"\4a" + "&HFF&"
            tag = "{" + tag + "}"

            event.Text = tag + event.Text
            return event

        def popup_box(event: Event) -> Event:
            event.Name = event.Name + "_popup_box"
            event.Layer = 0

            d = Draw()
            d.Add(Point(0, 0, "m"))
            d.Add(Point(width, 0, "l"))
            d.Add(Point(width, height, "l"))
            d.Add(Point(0, height, "l"))
            box = d.Dump()
            box = r"{\p1}" + box + r"{\p0}"

            tag = ""
            tag += r"\an7" + r"\pos({},{})".format(x, y)
            tag += r"\fs" + str(textSize)
            tag += r"\c" + bgColor
            tag += r"\1a" + bgOpacity
            tag += r"\2a" + "&HFF&" + r"\3a" + "&HFF&" + r"\4a" + "&HFF&"
            tag = "{" + tag + "}"

            event.Text = tag + box
            return event

        def title(event: Event) -> Event:
            # 很明显 title 字体大小是用 DPI 72 计算的
            textSize = round(each.textSize / 4, 3)

            event.Name += "_title"

            tag = ""
            tag += r"\an7" + r"\pos({},{})".format(x, y)
            tag += r"\fs" + str(textSize)
            tag += r"\c" + fgColor
            tag += r"\b500"
            tag += r"\2a" + "&HFF&" + r"\3a" + "&HFF&" + r"\4a" + "&HFF&"
            tag = "{" + tag + "}"

            event.Text = tag + event.Text
            return event

        events: List[Event] = []
        event = Event()

        event.Start = each.timeStart
        event.End = each.timeEnd
        event.Name = each.id

        text = each.text
        text = text.replace("\n", r"\N")
        if libass:
            # 仅 libass 支持大括号转义
            text = text.replace(r"{", r"\{")
            text = text.replace(r"}", r"\}")
        event.Text = text

        x = round(each.x, 3)
        y = round(each.y, 3)
        textSize = round(each.textSize, 3)
        fgColor = ConvertColor(each.fgColor)
        bgColor = ConvertColor(each.bgColor)
        bgOpacity = ConvertAlpha(each.bgOpacity)
        width = round(each.width, 3)
        height = round(each.height, 3)

        if libass:
            # 针对 libass 的 hack
            width = width * 1.776
        width = round(width, 3)

        event.Layer = 1

        if each.style == "popup":
            events.append(popup(copy.copy(event)))
            events.append(popup_box(copy.copy(event)))
        elif each.style == "title":
            events.append(title(copy.copy(event)))
        else:
            print(_("本脚本暂不支持 {} 样式. ({})").format(each.style, each.id))

        return events

    events = []
    for each in annotations:
        events.extend(ConvertAnnotation(each))

    return events



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


class Color(object):
    """以 0-255 的整数表示颜色值, 不提供序列化方法"""

    def __init__(
        self,
        red: int = 0,
        green: int = 0,
        blue: int = 0,
    ):
        self.red = red
        self.green = green
        self.blue = blue


class Alpha(object):
    """以 0-255 的整数表示透明度, 不提供序列化方法"""

    def __init__(
        self,
        alpha: int = 0,
    ):
        self.alpha = alpha


class Rgba(object):
    def __init__(self, color: Color = Color(), alpha: Alpha = Alpha()):
        self.color = color
        self.alpha = alpha
